/* DatabaseView.java
 *
 * Copyright 1998 Jim McBeath.
 *
 * June 7, 1998
 */

package jimmc.roots;

import jimmc.swing.ButtonAction;
import jimmc.swing.ComboBoxAction;
import jimmc.swing.GridBagger;
import jimmc.swing.MenuAction;
import jimmc.swing.JsTextField;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.awt.GridBagConstraints;
import java.io.PrintWriter;
import javax.swing.border.EtchedBorder;
import javax.swing.border.TitledBorder;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;

/** The DatabaseView window provides a view into the DatabaseModel object.
 */
public class DatabaseView extends RootsFrame {
	/** The database we are viewing. */
	protected DatabaseModel model;

	/** The selection keyword entry field. */
	protected ComboBoxAction selectWhatField;

	/** The selection match entry field. */
	protected ComboBoxAction selectMatchField;

	/** The command text entry field. */
	protected JsTextField commandTextField;

	/** The selection text entry field. */
	protected JsTextField selectTextField;

	/** The Search button. */
	protected ButtonAction searchButton;

	/** Where we display the path to the database. */
	protected JLabel pathField;

	/** Where we display the number of records in the database. */
	protected JLabel recordCountField;

	/** Where we record the log text to the screen. */
	protected ScrolledText logField;
	
	/** Create a new DatabaseObserver. */
	public DatabaseView() {
		super();
		setJMenuBar(createMenuBar());
		initForm();
		pack();
		addWindowListener();
		setTitle("roots");
	}

	/** Set the database we are viewing. */
	public void setDatabase(String dbpath) {
		PrintWriter pw =
			new PrintWriter(new ScrolledTextWriter(logField));
		DatabaseModel m = new DatabaseModel(dbpath,pw);
		setDatabaseModel(m);
		setTitle("roots "+dbpath);
	}

	/** Set the database we are viewing. */
	public void setDatabaseModel(DatabaseModel db) {
		model = db;
		searchButton.setEnabled(true);
		pathField.setText(model.getPath());
		recordCountField.setText(
			Integer.toString(model.getRecordCount()));
		pack();		/* get the size right */
		validate();
	}

	/** Create our menu bar. */
	protected JMenuBar createMenuBar() {
		JMenuBar mb = new JMenuBar();
		mb.add(createFileMenu());
		JMenu helpMenu = createHelpMenu();
		mb.add(helpMenu);
		//mb.setHelpMenu(helpMenu);
		return mb;
	}

	/** Create our File menu. */
	protected JMenu createFileMenu() {
		JMenu m = new JMenu("File");
		MenuAction mi;
		mi = new MenuAction("Open") {
			public void action() {
				processFileOpen();
			}
		};
		m.add(mi);
		mi = new MenuAction("Exit") {
			public void action() {
				processFileExit();
			}
		};
		m.add(mi);
		return m;
	}

	/** Create the body of our form. */
	protected void initForm() {
		GridBagger gb = new GridBagger(getContentPane());
		gb.gbc.gridwidth = GridBagConstraints.REMAINDER;
		gb.gbc.fill = GridBagConstraints.HORIZONTAL;
		gb.add(createInfoArea());
		gb.nextRow();
		gb.add(createCommandArea());
		gb.nextRow();
		gb.add(createSelectArea());
		gb.nextRow();
		gb.gbc.weightx = 1;
		gb.gbc.weighty = 1;
		gb.gbc.fill = GridBagConstraints.BOTH;
		gb.add(createLogArea());
	}

	/** Create the info area. */
	protected Component createInfoArea() {
		JPanel p = new JPanel();
		p.setBorder(new TitledBorder(new EtchedBorder(),"Database"));
		GridBagger gb = new GridBagger(p);
		gb.gbc.fill = GridBagConstraints.HORIZONTAL;
		gb.gbc.anchor = GridBagConstraints.WEST;
		gb.add(new JLabel("Path:"));
		gb.add(pathField=new JLabel("...."));
		gb.nextRow();
		gb.add(new JLabel("Records:"));
		gb.gbc.weightx = 1;
		gb.add(recordCountField=new JLabel("...."));
		return p;
	}

	/** Create the command area. */
	protected Component createCommandArea() {
		JPanel p = new JPanel();
		p.setBorder(new TitledBorder(new EtchedBorder(),"Command"));
		GridBagger gb = new GridBagger(p);
		gb.gbc.anchor = GridBagConstraints.WEST;
		gb.add(new JLabel("Command:"));
		gb.gbc.fill = GridBagConstraints.HORIZONTAL;
		gb.gbc.weightx = 1;
		commandTextField = new JsTextField(30) {
			public void action() {
				processCommand();
			}
		};
		commandTextField.setToolTipText(
		    "Type a command and press Enter; to get started, try help");
		gb.add(commandTextField);
		return p;
	}

	/** Create the selection area. */
	protected Component createSelectArea() {
		JPanel p = new JPanel();
		p.setBorder(new TitledBorder(new EtchedBorder(),"Select"));
		GridBagger gb = new GridBagger(p);
		gb.gbc.fill = GridBagConstraints.HORIZONTAL;
		gb.gbc.anchor = GridBagConstraints.WEST;
		gb.gbc.gridwidth = 1;

		selectWhatField = new ComboBoxAction() {
			public void action() {
				String what = getSelectedString();
				boolean enableMatch = 
					!(what.equals("freeform"));
				if (selectMatchField!=null)
					selectMatchField.setVisible(
						enableMatch);
			}
		};
		selectWhatField.addItem("name");
		selectWhatField.addItem("place");
		selectWhatField.addItem("key");
		selectWhatField.addItem("freeform");
		gb.add(selectWhatField);

	
		selectMatchField = new ComboBoxAction();
		selectMatchField.addItem("includes");
		selectMatchField.addItem("startsWith");
		selectMatchField.addItem("matches");
		gb.add(selectMatchField);

		selectTextField = new JsTextField(30) {
			public void action() {
				processSelect();
			}
		};
		gb.gbc.weightx = 1;
		gb.add(selectTextField);

		searchButton = new ButtonAction("Search") {
			public void action() {
				processSelect();
			}
		};
		if (model==null)
			searchButton.setEnabled(false);
		gb.add(searchButton);

		return p;
	}
	
	/** Get the selection text. */
	protected String getSelectionText() {
		String txt = selectTextField.getText();
		String what = selectWhatField.getSelectedString();
		if (what.equals("freeform"))
			return txt;
		String match = selectMatchField.getSelectedString();
		String sel = what+" "+match+" "+quoteString(txt);
		return sel;
	}

	/** Quote a string. */
	protected String quoteString(String txt) {
		return "'"+txt+"'";		/* TBD - make smarter! */
	}

	/** Create the Log area. */
	protected Component createLogArea() {
		JPanel p = new JPanel();
		p.setBorder(new TitledBorder(new EtchedBorder(),"Log"));
		GridBagger gb = new GridBagger(p);
		gb.gbc.fill = GridBagConstraints.BOTH;
		gb.gbc.anchor = GridBagConstraints.WEST;
		gb.gbc.weightx = 1;
		gb.gbc.weighty = 1;
		logField = new ScrolledText(500,300);
		gb.add(logField);
		return p;
	}

	/** Process the File->Open menu command. */
	protected void processFileOpen() {
		errorDialog("File->Open not yet implemented."); /* TBD */
	}

	/** Closing this window is the same as File->Exit. */
	protected void processClose() {
		processFileExit();
	}

	/** Process the File->Exit menu command. */
	protected void processFileExit() {
		if (confirmExit()) {
			System.exit(0);
		}
	}

	/** Ask if he really wants to exit from the program. */
	protected boolean confirmExit() {
		String msg = "Exit jimmc.roots?";
		String title = "Confirm";
		int n = JOptionPane.showConfirmDialog(this,msg,title,
				JOptionPane.YES_NO_OPTION);
		return (n==0);
	}

	/** Process a command. */
	protected void processCommand() {
		String cmd = commandTextField.getText();
		model.logCommand(cmd);
		model.doCommandToLog(cmd);
	}

	/** Process a selection. */
	protected void processSelect() {
		if (model==null)	//not yet initialized
			return;		//ignore it
		String txt = getSelectionText();
		model.select(txt);
		/* TBD - check for success of selection? */
		String[] selectedKeys = model.getSelectedKeys();
		if (selectedKeys==null || selectedKeys.length==0) {
			/* no items selected */
			/* TBD - status line? */
			errorDialog("No match");
		} else {
			PolyView p = new PolyView(model,selectedKeys);
			p.show();
		}
	}

	/** A main routine. */
	public static void main(String[] args) {
		if (args.length<1) {
			System.err.println("please specify a database file");
			System.exit(1);
		}
		String dbpath = args[0];
		if (args.length>1) {
			//parse remaining command line args
			if (args[1].equals("-c")) {
				PrintWriter syspw = new PrintWriter(System.out);
				DatabaseModel model =
					new DatabaseModel(dbpath,syspw);
				model.doCommand(syspw,args[2]);
				System.exit(0);
			}
			System.out.println("Unknown argument "+args[1]);
			System.exit(1);
		}
		DatabaseView dbv = new DatabaseView();
		dbv.show();
		dbv.setDatabase(args[0]);
	}
}

/* end */
