/* Copyright 1997 Jim McBeath
 *
 * Jim McBeath, February 1, 1997
 */

package jimmc.roots;

import java.io.OutputStream;
import java.io.PrintWriter;
import java.io.Writer;
import java.util.Vector;

/** Provides support for multi-column output, assuming a fixed-width font.
 * After instantiating, use the addColumn method to define the print columns
 * and their characteristics.  Use selectColumn to select a column,
 * then just call the normal PrintWriter methods such as println to add
 * data to that column.  Use flushColumns to get all columns back into
 * alignment, and use selectDirect to print directly to the output (e.g.
 * to print data that spans all columns).
 */
public class MultiColumnPrintStream extends PrintWriter {
	Vector columns;
	MultiColumn currentColumn;

	/** Create a new MultiColumnPrintStream.
	 * @param out the output stream
	 */
	public MultiColumnPrintStream(OutputStream out) {
		super(out);
		columns = new Vector();
		currentColumn = null;
	}

	/** Create a new MultiColumnPrintStream.
	 * @param out the output stream
	 */
	public MultiColumnPrintStream(Writer out) {
		super(out);
		columns = new Vector();
		currentColumn = null;
	}

	/** Add an output column.
	 * @param width the width of the column in print positions
	 * @param spacing the number of blank characters between this
	 *           column and the next column.
	 * @return the index of this column, to be passed to selectColumn.
	 *         The first column created is 0, the next is 1, etc.
	 */
	public synchronized int addColumn(int width, int spacing) {
		MultiColumn mc = new MultiColumn(this,width,spacing);
		columns.addElement(mc);
		return columns.size()-1;	/* index of newest column */
	}

	/** Select the column to which output is directed.
	 * @param col the column index as returned by addColumn.
	 */
	public void selectColumn(int col) {
		currentColumn = (MultiColumn)columns.elementAt(col);
	}
	/** Flush and deselect all columns.
	 * Output goes directly to the underlying stream.
	 */
	public void selectDirect() {
		flushColumns();
		currentColumn = null;
	}

	/** Flush out the data from all columns.  If any column has a partial
	 * line, a newline is added to that column.  Then all columns are
	 * output to the underlying stream.  If one column has more lines of
	 * data than other columns, then blank lines are output in those
	 * other columns until all columns are fully output.  This has the
	 * effect of vertically synchronizing or aligning all columns.
	 */
	public synchronized void flushColumns() {
		MultiColumn mc;

		for (int i=0; i<columns.size(); i++) {
			mc = (MultiColumn)columns.elementAt(i);
			mc.finishPartialLine();	/* make sure no partial lines */
		}
		while (anyColumnHasData()) {
			outputOneLine();
		}
	}
	/** Output as many complete lines from all columns as possible. */
	public synchronized void outputCompleteLines() {
		int n = countCompleteLines();
		while (n-->0)
			outputOneLine();
	}

	/* @return number of lines that are complete in all columns. */
	private int countCompleteLines() {
		int c=0;
		for (int i=0; i<columns.size(); i++) {
			MultiColumn mc = (MultiColumn)columns.elementAt(i);
			int cc = mc.countCompleteLines();
			if (i==0 || cc<c) {
				c = cc;  /* minimum number of complete lines */
				if (c==0)
					return 0;	/* empty column */
			}
		}
		return c;
	}
	/** See if any column has any data.
	 * @return true if any of our columns have any data in them.
	 */
	private boolean anyColumnHasData() {
		return anyColumnHasData(0);
	}
	/** See if any column starting after a given point has any data.
	 * @param colnum index of the first column to check
	 * @return true if that column or any past it has any data.
	 */
	public synchronized boolean anyColumnHasData(int col) {
		for (; col<columns.size(); col++) {
			MultiColumn mc = (MultiColumn)columns.elementAt(col);
			if (mc.hasData()) {
				return true;	/* found some data */
			}
		}
		return false;	/* no data */
	}

	/* Output one line from all columns.
	 */
	private void outputOneLine() {
		for (int i=0; i<columns.size(); i++) {
			MultiColumn mc = (MultiColumn)columns.elementAt(i);
			boolean needFill = anyColumnHasData(i+1);
			mc.writeLine(needFill);
		}
		super.write(012);	/* newline */
	}

	/** Flush all columns, then flush the underlying stream. */
	public void flush() {
		flushColumns();
		super.flush();
	}

	/** Conditional println: print text to the output,
	 * add a newline if the last character is not already a newline. */
	public void printlnc(String s) {
		print(s);
		if (!s.endsWith("\n"))
			println();
	}

	/** Write a char to the currently selected column.
	 * Data is buffered in memory until a newline character is
	 * written, at which time output to the underlying stream may
	 * occur (or may not, depending on whether all columns have
	 * data).  If output does occur, the write may block.
	 * @param b the char
	 */
	public void write(int b) {
		if (currentColumn==null)
			super.write(b);
		else
			currentColumn.write(b);
	}

	/** Write a sub array of chars to the currently selected column.
	 * Equivalent to calling <a href="#write(int)">write(int)</a>
	 * for each char in the array.
	 * @param b the data to be written
	 * @param off the start offset in the data
	 * @param len the number of chars to write
	 */
	public void write(char b[], int off, int len) {
		if (currentColumn==null)
			super.write(b,off,len);
		else
			currentColumn.write(b,off,len);
	}

	/** Write a substring to the currently selected column.
	 * Equivalent to calling <a href="#write(int)">write(int)</a>
	 * for each char in the substring.
	 * @param s the string to be written
	 * @param off the start offset in the string
	 * @param len the number of chars to write
	 */
	public void write(String s, int off, int len) {
		if (currentColumn==null)
			super.write(s,off,len);
		else
			currentColumn.write(s,off,len);
	}

	/* Write a String to our underlying stream. */
	void writeDirect(String s) {
		super.write(s,0,s.length());
	}
	/* Write a char to our underlying stream. */
	void writeDirect(int b) {
		super.write(b);
	}
}

/* MultiColumn represents a single column in MultiColumnPrintStream. */
class MultiColumn {
	MultiColumnPrintStream ps;
	int width;
	int spacing;
	Vector completeLines;	/* vector of Strings */
	StringBuffer currentLine;

	MultiColumn (MultiColumnPrintStream ps, int width, int spacing) {
		this.ps = ps;
		this.width = width;
		this.spacing = spacing;
		completeLines = new Vector();
		currentLine = new StringBuffer();
	}

	/* Write a char into the column.  If the char is a newline,
	 * check to see if we can output anything. */
	public void write(int b) {
		if (b==012) {
			/* A newline, advance to the next line. */
			String s = currentLine.toString();
			completeLines.addElement(s);
			currentLine = new StringBuffer();
			/* Now see if we can output any lines. */
			ps.outputCompleteLines();
		} else {
			if (currentLine.length()>=width)
				write(012);	/* wrap to next line */
			currentLine.append((char)b);
		}
	}

	/** Write an array of chars into the column. */
	public void write(char b[], int off, int len) {
		while (len-->0) {
			write(b[off++]);
		}
	}

	/** Write an substring into the column. */
	public void write(String s, int off, int len) {
		s = s.substring(off,off+len);
		for (int i=0; i<len; i++) {
			int c = s.charAt(i);
			write(c);
		}
	}

	/** Write out a line of text.
	 * @param fill output spaces to fill the column width plus spacing
	 */
	public synchronized void writeLine(boolean fill) {
		String s;
		if (completeLines.size()==0) {
			if (currentLine.length()==0) {
				s = ""; /* nothing to write */
			} else {
				s = currentLine.toString();
				currentLine = new StringBuffer();
			}
		} else {
			s = (String)(completeLines.elementAt(0));
			completeLines.removeElementAt(0);
		}
		ps.writeDirect(s);
		if (fill) {
			int n = width+spacing - s.length(); /* num fill chars */
			while (n>0) {
				ps.writeDirect(' '); /* write a space */
				n--;
			}
		}
	}

	/** If there is a partial line in this column, add a newline. */
	public void finishPartialLine() {
		if (currentLine.length()!=0) {
			write(012);	/* write a newline */
		}
	}

	/** @return the number of complete lines in this column. */
	public int countCompleteLines() {
		return completeLines.size();
	}

	/** @return true if this column has any data in it
	 *         (full or partial lines).
	 */
	public boolean hasData() {
		return (currentLine.length()>0 || completeLines.size()>0);
	}
}

/* end */
