/* RootsBase.java
 * Copyright 1997 Jim McBeath
 *
 * Jim McBeath, January 12, 1997
 */

package jimmc.roots;

import jimmc.treepar.Layout;

import java.io.*;

/** RootsRecord is the generic record class from which all other record
 * classes are derived.
 */
public abstract class RootsRecord {
	protected RootsBase base;
	protected String key;		/* unique key for this data record */
	protected String keyComment;	/* stuff on key line after the key */
	protected String editDate;	/* date this record was last edited */
	protected String note1;		/* short note */
	protected String note2;		/* medium note */
	protected String note2d;	/* delta medium note (add to short) */
	protected String comment; /* not output anywhere, only for browsing */
	protected String sourceString;
	protected RootsSource sources[]; /* sources of this data record */

	public RootsRecord(RootsBase base) {
		this.base = base;
	}

	/** Get the key for this record. */
	public String getKey() {
		return key;
	}

	/* Returns true if we recognized the name and added the value */
	protected boolean addNcv(Ncv nv) {
		if      (nv.name.equals("EDITED"))
			editDate = nv.value;
		else if (nv.name.equals("NOTE1"))
			note1 = nv.value;
		else if (nv.name.equals("NOTE2"))
			note2 = nv.value;
		else if (nv.name.equals("NOTE2D"))
			note2d = nv.value;
		else if (nv.name.equals("COM"))
			comment = nv.value;
		else if (nv.name.equals("SOURCE"))
			sourceString = nv.value;
		else
			return false;	/* not recognized */
		return true;		/* OK */
	}

	/** Fix up the links in a record that point to another record. */
	protected void link() {
		RootsSource sa[];
		if ((sa=RootsSource.FindRecordList(base,sourceString))!=null)
			sources = sa;
	}

	/** Check and parse any dates in this record into individual fields. */
	protected void parseDates() {
		/* default is to do nothing */
	}

	/** Parse a single date, give error message if bad. */
	protected void parseDate(RootsEvent d, String m) {
		if (!d.dateToParts()) {
			println("Bad "+m+" date in record "+key);
		}
		d.dateFromParts();	/* normalize it */
	}

	/** Format this record as GEDCOM. */
	public String formatGedcom() {
		StringBuffer sb = new StringBuffer();
		formatGedcomKey(sb,0);
		formatGedcomBody(sb,1);	/* body is one level down from key */
		return sb.toString();
	}

	/** Format the key line as GEDCOM.
	 * Subclass must override to make this work.
	 */
	protected void formatGedcomKey(StringBuffer sb, int level) {
		/* do nothing here. */
	}

	/** Format the body of the record.
	 */
	protected void formatGedcomBody(StringBuffer sb, int level) {
		/* TBD - which note to output? */
		formatGedcomItem(sb,level,"NOTE",note1);
	}

	/** Format the key line into the specified buffer.
	 */
	protected void formatGedcomKey(StringBuffer sb, int level, String tag) {
		sb.append(level).append(" @");
		sb.append(key).append("@ ").append(tag).append("\n");
	}

	/** Format a reference into the specified buffer.
	 */
	protected void formatGedcomReference(StringBuffer sb, int level,
				String tag, RootsRecord ref) {
		if (ref==null || ref.key==null)
			return;
		sb.append(level).append(" ").append(tag);
		sb.append(" @").append(ref.key).append("@\n");
	}
	protected void formatGedcomItem(StringBuffer sb, int level,
				String tag, String value) {
		if (value==null)
			return;
		sb.append(level).append(" ").append(tag).append(" ");
		int n = value.indexOf("\n");
		if (n>=0)
			formatMultiLine(sb,value,(level+1)+" CONT ");
		else
			sb.append(value);
		sb.append("\n");
	}
	protected void formatGedcomEvent(StringBuffer sb, int level,
				String tag, RootsEvent ev) {
		if (ev==null)
			return;
		if (ev.date==null && ev.place==null)
			return;
		sb.append(level).append(" ").append(tag).append("\n");
		if (ev.date!=null) {
			sb.append(level+1).append(" DATE ");
			sb.append(ev.date).append("\n");
		}
		if (ev.place!=null) {
			sb.append(level+1).append(" PLAC ");
			sb.append(ev.place).append("\n");
		}
	}

	/** Format this record as NCV. */
	public String formatNcv() {
		StringBuffer sb = new StringBuffer();
		sb.append(formatKeyLine());
		sb.append("\n");
		sb.append(formatNcvType());
		formatNcvBody(sb);
		return sb.toString();
	}
	/** Return the key line for this record */
	public String formatKeyLine() {
		return (key+keyComment);
	}
	/** Format the type of this record; subclass must override */
	public abstract String formatNcvType();
	/** Format the remainder of this record */
	public void formatNcvBody(StringBuffer sb) {
		formatNcvItem(sb,"NOTE1",note1);
		formatNcvItem(sb,"NOTE2",note2);
		formatNcvItem(sb,"NOTE2D",note2d);
		formatNcvItem(sb,"COM",comment);
		formatNcvItem(sb,"SOURCE",sourceString);
		formatNcvItem(sb,"EDITED",editDate);
	}
	protected void formatNcvItem(StringBuffer sb,String label,String value){
		if (value==null || value.equals(""))
			return;	/* nothing to output */
		sb.append("\n").append(label).append(":");
		formatMultiLine(sb,value,"+");
	}
	protected void formatMultiLine(StringBuffer sb, String value,
					String linePrefix) {
		int n = value.indexOf("\n");
		while (n>=0) {
			/* The string contains newlines, we need to split at
			 * each line and add a "+" at the start of each line. */
			String s = value.substring(0,n+1);
			sb.append(s);
			value = value.substring(n+1);
			if (value.length()>0)
				sb.append(linePrefix);
			n = value.indexOf("\n");
		}
		sb.append(value);
	}
	/** Return a key string */
	public String formatKey() {
		StringBuffer sb = new StringBuffer();
		formatKey(sb);
		return sb.toString();
	}
	/** Write the key into the buffer */
	public void formatKey(StringBuffer sb) {
		sb.append("[").append(key).append("] ");
	}
	/** Given a label and one value, print them to the StringBuffer
	 * if the value is not null.  If null, print nothing. */
	protected final void formatLabeledItem(StringBuffer sb,
				String label, String value) {
		formatLabeledItem(sb,label,value,"");
	}
	protected final void formatLabeledItem(StringBuffer sb,
				String label, String value, String after) {
		if (value!=null && !value.trim().equals(""))
			sb.append(label).append(value).append(after);
	}
	/** Given a label and two values, print them to the StringBuffer
	 * if either is not null.  Also print a trailing space.
	 * If both null, print nothing. */
	protected final void formatLabeledItems(StringBuffer sb,
				String label, String value1, String value2) {
		formatLabeledItems(sb,label,value1,value2,"");
	}
	protected final void formatLabeledItems(StringBuffer sb,
		     String label, String value1, String value2, String after) {
		if ((value1!=null && !value1.trim().equals("")) ||
		    (value2!=null && !value2.trim().equals(""))) {
			sb.append(label);
			if (value1!=null && !value1.trim().equals(""))
				sb.append(value1).append(" ");
			if (value2!=null && !value2.trim().equals(""))
				sb.append(value2).append(" ");
			sb.append(after);
		}
	}
	protected final void formatNote1(StringBuffer sb,
				String before, String after) {
		if (note1==null)
			return;
		sb.append(before).append(note1).append(after);
	}

	/** Format a one-line summary. */
	public abstract String formatOneLine(PrintOptions po);
	/** Format a brief listing of the record */
	public String formatBrief(PrintOptions po) {
		/* If not overridden, just output one line */
		return formatOneLine(po);
	}
	/** Format a medium listing of the record */
	public String formatMedium(PrintOptions po) {
		/* If not overridden, output brief listing */
		return formatBrief(po);
	}

	/** Get the keys for the records used to generate a ctree
	 * with this record as the focus.
	 */
	public String[] getCtreeReferenceKeys() {
		return null;		//no keys by default
	}

	/** Add our Ctree information to the layout for this record
	 * and for all other records required for the ctree layout
	 * using this record as the root. */
	public void layoutCtree(Layout layout, PrintOptions po) {
		/* Do nothing by default */
	}

	/** Add a single box to the layout for this record. */
	protected void layoutBox(Layout layout, PrintOptions po, 
						String markKey) {
		/* no default */
	}

	/** Add boxes for the referenced items. */
	protected void layoutBoxReference(Layout layout, PrintOptions po,
					String reference) {
		RootsRecord[] rr = findReference(reference);
		if (rr==null)
			return;		/* nothing found */
		for (int i=0; i<rr.length; i++)
			rr[i].layoutBox(layout,po,null);
	}

	/** Find a referenced record.
	 * @param what the name of the reference item to follow
	 * @return the referenced record, or null if not found */
	public RootsRecord[] findReference(String what) {
		if (what==null)
			return null;
		if (what.startsWith("source"))
			return findListReference(sources,what.substring(6));
		return null;
	}

	/** Given a string as a number, find the item in a list for it.
	 * @param list the list of records
	 * @param numstr the index number N as a string (first item is 0)
	 * @return the Nth item in the list
	 */
	public RootsRecord[] findListReference(
				RootsRecord list[], String numstr){
		if (list==null)
			return null;	/* no list to reference */
		if (numstr.equals(""))
			return list;	/* no index, return the whole list */
		int num;
		num = Integer.parseInt(numstr);
		if (num<0 || num>=list.length)
			return null;	/* not that many */
		return list[num].asArray();
	}

	/** Return true if the specified field matches the given pattern.
	 * @param fieldName the name of the field to match.
	 * @param the pattern to match against.
	 */
	public boolean fieldMatches(String fieldName, String pattern) {
		/* subclass should override. */
		if (fieldName.equals("key")) {
			return (PatternMatch.stringMatches(key,pattern));
		}
		if (fieldName.equals("note"))
			return PatternMatch.stringMatches(note1,pattern) ||
				PatternMatch.stringMatches(note2,pattern) ||
				PatternMatch.stringMatches(note2d,pattern);
		return false;	/* unknown or unimplemented field */
	}

	/** Return an array referencing this.
	 * @return an array of length one with this object referenced
	 */
	public RootsRecord[] asArray() {
		RootsRecord ra[] = new RootsRecord[1];
		ra[0] = this;
		return ra;
	}

	/** Print a line of text to the output stream associated with our
	 * base.
	 */
	public void println(String s) {
		base.println(s);
	}
}

/* end */
