/* ArrayUtil.java
 *
 * Jim McBeath, October 25, 2001
 */

package jimmc.util;

import java.util.Arrays;
import java.util.Random;
import java.util.Vector;

/** Utility methods for arrays. */
public class ArrayUtil {
	/** Concatenate two string arrays into a new string array. */
	public static String[] cat(String[] a1, String[] a2) {
		if (a1==null)
			return a2;
		if (a2==null)
			return a1;
		String[] aa = new String[a1.length+a2.length];
		System.arraycopy(a1,0,aa,0,a1.length);
		System.arraycopy(a2,0,aa,a1.length,a2.length);
		return aa;
	}

	/** Take the union of two unsorted sets of strings. */
	public static String[] union(String[] a1, String[] a2) {
		Vector vv = new Vector();
		for (int i=0; i<a1.length; i++) {
			vv.addElement(a1[i]);
		}
		for (int i=0; i<a2.length; i++) {
			if (!vv.contains(a2[i]))
				vv.addElement(a2[i]);
		}
		String[] aa = new String[vv.size()];
		vv.copyInto(aa);
		return aa;
	}

	/** Given two sorted arrays of strings, return the list of all
	 * entries which are in the first array and not in the second.
	 * @param a1 All returned entries are from this list.
	 * @param a2 No returned entries are in this list.
	 * @return All entries in a1 and not in a2.
	 *         If a1 is null, returns null.
	 *         If a2 is null, returns a1.
	 */
	public static String[] andNot(String[] a1, String[] a2) {
		if (a1==null)
			return null;
		if (a2==null)
			return a1;
		Vector v = new Vector();
		int i1 = 0;
		int i2 = 0;
		while (i1<a1.length && i2<a2.length) {
			int x = a1[i1].compareTo(a2[i2]);
			if (x>0) {
				//Need to bring a2 along
				i2++;
			} else if (x==0) {
				//The strings match, don't include it in result
				i1++;
				i2++;
			} else /* x<0 */ {
				//a2 is already past here, so it's not in a2
				v.addElement(a1[i1]);
				i1++;
			}
		}
		while (i1<a1.length) {
			//Ran out of a2, put everything into the results
			v.addElement(a1[i1]);
			i1++;
		}
		String[] r = new String[v.size()];
		v.copyInto(r);
		return r;
	}

	/** Given two sorted arrays of ints, return the list of all
	 * entries which are in the first array and not in the second.
	 * @param a1 All returned entries are from this list.
	 * @param a2 No returned entries are in this list.
	 * @return All entries in a1 and not in a2.
	 *         If a1 is null, returns null.
	 *         If a2 is null, returns a1.
	 */
	public static int[] andNot(int[] a1, int[] a2) {
		if (a1==null)
			return null;
		if (a2==null)
			return a1;
		int i1 = 0;
		int i2 = 0;
		boolean[] a1marks = new boolean[a1.length];
		int a1count = 0;	//number of marks in a1marks
		while (i1<a1.length && i2<a2.length) {
			int x = a1[i1]-a2[i2];
			if (x>0) {
				//Need to bring a2 along
				i2++;
			} else if (x==0) {
				//The numbers match, don't include it in result
				i1++;
				i2++;
			} else /* x<0 */ {
				//a2 is already past here, so it's not in a2
				a1marks[i1] = true;
				a1count++;
				i1++;
			}
		}
		while (i1<a1.length) {
			//Ran out of a2, put everything into the results
			a1marks[i1] = true;
			a1count++;
			i1++;
		}
		int[] r = new int[a1count];
		int rx = 0;
		for (i1=0; i1<a1.length; i1++) {
			if (a1marks[i1])
				r[rx++] = a1[i1];
		}
		return r;
	}

	/** Get an array of randomly ordered ints.
	 * @param n The size of the array.
	 * @return An array of length n, with entries from 0 to n-1 in
	 *         random order.
	 */
	public static int[] randomArray(int n) {
		Random rand = new Random();
		long[] nn = new long[n];
		for (int i=0; i<n; i++) {
			long r = rand.nextInt();
			nn[i] = ((r<<32)&0x7FFFFFF)+i;
		}
		Arrays.sort(nn);
		int[] results = new int[n];
		for (int i=0; i<n; i++) {
			results[i] = (int)(nn[i]&0x7FFFFFFF); //the int portion
		}
		return results;
	}

	/** Shuffle an array in random order. */
	public static Object[] shuffle(Object[] aa) {
		if (aa==null)
			return null;
		int n = aa.length;
		int[] randomIndexes = randomArray(n);
		Object[] results = new Object[n];
		for (int i=0; i<n; i++) {
			int x = randomIndexes[i];
			results[i] = aa[x];
		}
		return results;
	}

	/** Shuffle an array of strings in place. */
	public static void shuffle(String[] ss) {
		if (ss==null)
			return;
		Object[] rr = shuffle((Object[])ss);
		for (int i=0; i<rr.length; i++)
			ss[i] = (String)(rr[i]);
	}
}

/* end */
