/* Duration.java
 *
 * Jim McBeath, November 7, 2001
 */

package jimmc.util;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.math.BigDecimal;

//At first I made Duration a subclass of BigDecimal, but then when it gets
//passed to messageFormat it tries to format it as a number.

/** A Duration is an elapsed time measured in seconds. */
public class Duration implements Serializable {
	/** Our value. */
	protected BigDecimal duration;

	/** The number of fractional digits to display. */
	protected int fractionPrecision;

	/** Create a Duration from an int.
	 * The fractionPrecision is set to 0.
	 */
	public Duration(int duration) {
		this.duration = new BigDecimal((double)duration);
		setFractionPrecision(0);
	}

	/** Create a Duration from a double.
	 * The fractionPrecision is set to 2.
	 */
	public Duration(double duration) {
		this.duration = new BigDecimal(duration);
		setFractionPrecision(2);
	}

	/** Create a Duration from any kind of Number.
	 * If the number is a Float, Double, or BigDecimal, the
	 * fractionPrecision is set to 2; otherwise, it is set to 0.
	 */
	public Duration(Number num) {
		if (num==null) {
			duration = new BigDecimal(0.0);
			return;
		}
		duration = new BigDecimal(num.doubleValue());
		if ((num instanceof Double)||
		    (num instanceof Float)||
		    (num instanceof BigDecimal)) {
			setFractionPrecision(2);
		}
	}

	/** Set the number of fractional digits to display. */
	public void setFractionPrecision(int digits) {
		fractionPrecision = digits;
	}

	/** Get the number of fractional digits to display. */
	public int getFractionPrecision() {
		return fractionPrecision;
	}

	/** Get the value of this duration as an int.
	 */
	public int intValue() {
		return duration.intValue();
	}

	/** Get the value of this duration as a float. */
	public float floatValue() {
		return duration.floatValue();
	}

	/** Get the fractional part of our duration. */
	public double fractionalValue() {
		double d = duration.doubleValue();
		return d - Math.floor(d);
	}

	/** Convert to a string. */
	public String toString() {
		StringBuffer sb = new StringBuffer();
		int seconds = intValue();
		toString(sb,seconds);

		if (fractionPrecision>0) {
			sb.append(".");
			double f = fractionalValue();
			f += 0.5*Math.pow(10.0,-fractionPrecision);
				//add a rounding factor just past the last digit
			for (int i=0; i<fractionPrecision; i++) {
				f = f*10;
				int d = ((int)f);
				sb.append(Integer.toString(d));
				f -= d;
			}
		}
		return sb.toString();
	}

	/** Convert seconds to a Duration string. */
	public static String toString(int seconds) {
		StringBuffer sb = new StringBuffer();
		toString(sb,seconds);
		return sb.toString();
	}

	/** Convert seconds to string in a buffer. */
	protected static void toString(StringBuffer sb, int seconds) {
		int minutes = seconds/60;
		seconds = seconds%60;
		int hours = minutes/60;
		minutes = minutes%60;
		int days = hours/24;
		hours = hours%24;
		if (days>0) {
			sb.append(Integer.toString(days));
			sb.append("d ");
		}
		if (days>0 || hours>0) {
			sb.append(Integer.toString(hours/10));
			sb.append(Integer.toString(hours%10));
			sb.append(":");
		}
		sb.append(Integer.toString(minutes/10));
		sb.append(Integer.toString(minutes%10));
		sb.append(":");
		sb.append(Integer.toString(seconds/10));
		sb.append(Integer.toString(seconds%10));
	}

	/** Parse a string into an integer-valued duration.
	 * @return The number of seconds in the duration.
	 */
	public static int parseIntDuration(String s) {
		int days = 0;
		int dPos = s.indexOf('d');
		if (dPos>0) {
			String dayStr = s.substring(0,dPos);
			s = s.substring(dPos+1).trim();
			days = Integer.parseInt(dayStr);
		}
		String[] cStrs = StringUtil.toArray(s,':');
		int hours = 0;
		int minutes = 0;
		int seconds = 0;
		String msg;
		switch (cStrs.length) {
		case 0:
			msg = "No time in duration string";
				//TBD - i18n
			throw new NumberFormatException(msg);
		case 1:
			msg = "No colons in duration string";
				//TBD - i18n
			throw new NumberFormatException(msg);
		case 2:		//minutes and seconds; no hours
			if (days>0) {
				msg = "Days but no hours in duration";
					//TBD - i18n
				throw new NumberFormatException(msg);
			}
			minutes = Integer.parseInt(cStrs[0]);
			seconds = Integer.parseInt(cStrs[1]);
			break;
		case 3:		//hours, minutes, and seconds
			hours = Integer.parseInt(cStrs[0]);
			minutes = Integer.parseInt(cStrs[1]);
			seconds = Integer.parseInt(cStrs[2]);
			break;
		default:	//too big
			msg = "Too many colons in duration string";
				//TBD - i18n
			throw new NumberFormatException(msg);
		}
		int n = ((days*24+hours)*60+minutes)*60+seconds;
		return n;
	}

	/** Parse a string into a double-valued duration.
	 * @return The number of seconds in the duration.
	 */
	public static double parseDoubleDuration(String s) {
		int dotPos = s.indexOf('.');
		if (dotPos<0) {
			//No fractional part, parse it as an int
			return (double)parseIntDuration(s);
		}
		int n = parseIntDuration(s.substring(0,dotPos));
		double f = parseFraction(s.substring(dotPos+1));
		return ((double)n)+f;
	}

	/** Parse a fractional number.
	 * @param s The fractional part, starting with the digit just after
	 *          the decimal point.  Digits only.
	 * @return The fractional part, between 0 inclusive and 1 exclusive.
	 */
	protected static double parseFraction(String s) {
		int n = Integer.parseInt(s);
		int len = s.length();
		return ((double)n)/Math.pow(10.0,(double)len);
	}

	/** Return a new Duration initialized to the value
	 * of the specified String.
	 */
	public static Duration valueOf(String s) {
		return new Duration(parseIntDuration(s));
	}

    //The Serializable interface
	private void writeObject(ObjectOutputStream s) throws IOException {
		s.writeObject(duration);
		s.writeInt(fractionPrecision);
	}

	private void readObject(ObjectInputStream s) 
			throws IOException, ClassNotFoundException {
		duration = (BigDecimal)(s.readObject());
		setFractionPrecision(s.readInt());
	}
    //End Serializable interface
}

/* end */
